//#include <Sodaq_MQTT_Interface.h>
//#include <Sodaq_R4X_MQTT.h>
//#include <Sodaq_R4X.h>

#include <PubSubClient.h>
////#include <Wire.h>
#include <TimeLib.h>
////#include <RTCZero.h>
#include <RTC_A.h>
#include <MKRNB.h>
////#include <ArduinoLowPower.h>
#include "MAX30105.h" 
#include <Adafruit_Si7021.h>

//#define ARDUINO_SODAQ_SARA
#define LOG true
#define l(x) if (LOG) DEBUG_STREAM.println(x);
#define DELAY (15 * 1000)
#define TIMEOUT (5 * 1000)
#define SEALEVELPRESSURE_HPA (1013.25)
#define TOKEN ""

//error defines
//Uses dumby data if false
#define _CHECK_SENSORS false
//prints debug info
#define _DEBUG true
////debuf versions of print and println to easily remove debug statements
////#if _DEBUG
#define DEBUG_PL(x) if(_DEBUG) DEBUG_STREAM.println(x);
#define DEBUG_P(x) if(_DEBUG) DEBUG_STREAM.print(x);
////#else
////#define DEBUG_PL(x) 
////#define DEBUG_P(x) 
////#endif
////pins to display if errors occured, may be used to turn on an LED if there is an error
#define MAX30105_ERROR_PIN P3_1
#define SI7021_ERROR_PIN P2_6
#define GPRS_ERROR_PIN P2_3
#define NBACCESS_ERROR_PIN P_8_1
#define MQTT_ERROR_PIN P3_0
//used to indicate if there is a problem
#define INDICATE_ERROR(x) pinMode(x,OUTPUT); digitalWrite(x,HIGH);
#define INDICATE_ERROR_I(x) pinMode(x,OUTPUT); digitalWrite(x,LOW);

// RTC Vars
// Real time clock variable, used to save the time of each data gathering cycle when modem isnt on
time_t current_time = 0;
//NB Modem Vars
const char PINNUMBER[]     = "3866"; //for sim card
#define MODEM_STREAM Serial1
#define DEBUG_STREAM Serial
NBClient nbClient;
GPRS gprs;
NB nbAccess;

//Sensor Vars
MAX30105 particleSensor;
//Create Instance of HTU21D or SI7021 temp and humidity sensor and MPL3115A2 barrometric sensor
Adafruit_Si7021 sensor = Adafruit_Si7021();

float humidity = 0;
float tempf = 0;
float tempc = 0;
int Si_temp = 0;
int Si_humi = 0;
//int MAX30105_fire_percentage = 0;
int MAX30105_temp = 0;
int MAX30105_Green = 0;
int MAX30105_Red = 0;
int MAX30105_IR = 0;
char time_string[25] = "0000-00-00T00:00:00";

// how often measurements should be taken in ms (external interupt can still trigger data gathering)
#define MEASUREMENT_INTERVAL_MS 30*1000
#define TRANSMISSION_INTERVAL_MS 24*60*60*1000
//amount of values to be queued up between a full cycle
//a cycle is when the data should be transmitted (other than when the external interrupt is triggered)
//Eg, cycle time = storage length * measurement interval
#define STORAGE_LEN 2//2*6*24 // (measures_per_10min * 10min_in_hr * hr_in_day)
//external interrupt pin
#define INTERRUPT_PIN 0
//temperature values to have a premature transmission. If the temperatures are reached all current data will be sent.
//temperature is in Celcius
#define MAX30105_TEMP_TRIGGER 60
#define SI7021_TEMP_TRIGGER 60
#define FILL_ARR(arr,val) for(int i = 0;i<STORAGE_LEN;i++){arr[i] = val;}
int dummyVal = 0;
//variables to store the data of each measurement interval
//The storage arrays act like an array that is resized from 0 to the STORAGE_LEN:
//  The current position on the array is stored to act as the current end of the "resizable list" (could be seen as making the list one item larger).
//  The length of the "resizable list" is saved to "data_obtained" whenever the current position of the list is set to zero (when its set to zero its similar to the list being reset to an empty list).
//  The recent data is added to the "end" of this "resized" list.
//  This method makes memory use constant and limits the number of rewrites of memory that must occur:
//    Instead of remaking or emptying the array by writing in each spot, the old data is just overwritten with the "list length" being the length of any new data
int data_arr_ind = 0;
int data_obtained = 0;

int Si_temp_arr[STORAGE_LEN] __attribute__ ((section (".noinit")));
int Si_humi_arr[STORAGE_LEN] __attribute__ ((section (".noinit")));
int MAX30105_temp_arr[STORAGE_LEN] __attribute__ ((section (".noinit")));
int MAX30105_Green_arr[STORAGE_LEN] __attribute__ ((section (".noinit")));
int MAX30105_Red_arr[STORAGE_LEN] __attribute__ ((section (".noinit")));
int MAX30105_IR_arr[STORAGE_LEN] __attribute__ ((section (".noinit")));
//char time_string_arr[STORAGE_LEN][25];
time_t time_arr[STORAGE_LEN] __attribute__ ((section (".noinit")));

//MQTT Vars
//contains all values for the MQTT client
#define MSG_LEN 256
#define MQTT_RETRIES 100
const int mqtt_port = 1883;
const char* mqtt_server = "138.130.173.160";//"mttq.cumulocity.com";
const char* mqtt_pass = "Detroit313";
const char* mqtt_user = "jcu/julius.arnoldjanco@my.jcu.edu.au";
const char* mqtt_name = "Fire_Alarm_TEST_MSP430";
char msg[MSG_LEN];
int msg_length = 0;
PubSubClient client(nbClient);

//Value Storage Vars
//Used for making the program know that the external interrupt was triggered
volatile bool interrupt_triggered = false;


//
//bool modemConnect(){
//    if (r4x.isConnected()) {
//        return true;
//    }
//    else {
//        r4x.setPin(PINNUMBER);
//        return r4x.connect(APN, URAT, MNOPROF, OPERATOR, M1_BAND_MASK, NB1_BAND_MASK);
//    }
//}


//function that just locks the program up, used if a critical error occurs
void system_failure(){
    
    while(1){
      digitalWrite(RED_LED, HIGH);
      delay(1000);
      digitalWrite(RED_LED, LOW);
      delay(1000);
    }
}



//used to shutdown the NB modem, "wait" makes the function wait until the NB modem is shutdown 
void shutdown_modem(bool wait){
    DEBUG_PL("Shutting down NB modem.")
    gprs.detachGPRS(wait);
    DEBUG_PL("GPRS Detached.");
    if(wait){
        nbAccess.shutdown();
        DEBUG_PL("NBAccess close finished.");
        return;
    }
    nbAccess.secureShutdown();
    DEBUG_PL("NBAccess close finished.");
    MODEM_STREAM.end();
}


void setup_modem(){
    #if _DEBUG
    Serial.begin(115200);
    unsigned long tmp_time = millis();
    #endif
    DEBUG_PL("Setting up NB modem.")
    boolean connected = false;
    NB_NetworkStatus_t nb_status = ERROR;
    while (!connected) {
        //starts up the NB modem
        if (nb_status != NB_READY){
            nb_status = nbAccess.begin(PINNUMBER);
            if(nb_status == NB_READY){
              DEBUG_PL("NB Access granted.");
            }
        }else{
            DEBUG_PL("NB Access not granted.");
        }
        //after the NB modem is started and is connected, attach the GPRS
        if(nb_status == NB_READY){
            if(gprs.attachGPRS() == GPRS_READY) {
                connected = true;
                DEBUG_PL("GPRS Access granted.");
            }else {
                DEBUG_PL("Not connected");
                delay(100);
            }
        } 
    }
    #if _DEBUG
    DEBUG_PL(nbAccess.__loops);
    DEBUG_PL(nbAccess.__time);
    DEBUG_PL(nbAccess.__time_modem);
    DEBUG_PL(MODEM.__time_autosense);
    DEBUG_PL(MODEM.__time_reset);
    DEBUG_PL(MODEM.__time_uart);
    DEBUG_PL(MODEM.__time_send);
    DEBUG_PL(gprs.__time);
    DEBUG_P("NB TIME: ");
    DEBUG_PL((float)(millis()-tmp_time));
    #endif
}
//shuts down sensors
void shutdownSensors(){
  particleSensor.shutDown();
  //digitalWrite(13,LOW);
}


//used to setup the sensors
void setup_sensors(){
    //digitalWrite(9,HIGH);
    if(!_CHECK_SENSORS){
      DEBUG_PL("Skipping sensor setup.")
      return;
    }
    bool error = false;
    DEBUG_PL("Setting up MAX30105 Sensor.")
    if (particleSensor.begin(Wire, I2C_SPEED_FAST) == false) //Use default I2C port, 400kHz speed
    {
        DEBUG_PL("MAX30105 was not found. Please check wiring/power. ")
        INDICATE_ERROR(MAX30105_ERROR_PIN)
        error = true && _CHECK_SENSORS;
        
    }
    particleSensor.setup();
    DEBUG_PL("Setting up SI7021 Sensors.")
    //digitalWrite(13,HIGH);
    int ret = sensor.begin();
    //ret is 1 if setup has worked
    if(ret != 1){
        DEBUG_P("SI7021 not found. (")
        DEBUG_P(ret)
        DEBUG_PL(")")
        INDICATE_ERROR(SI7021_ERROR_PIN)
        error = true && _CHECK_SENSORS;
    }
    //if an error occured, make sure modem is shutdown and try setup sensors again. May add a system failure instead of it trying again
    if(error){
        shutdown_modem(false);
        delay(1000);
        system_failure();
    }
    DEBUG_PL("Finishing sensor setup.");
    //digitalWrite(9,LOW);
}


////fills the storage arrays with default values
//void setup_vars(){
////  FILL_ARR(Si_temp_arr,0)
////  FILL_ARR(Si_humi_arr,0)
////  //FILL_ARR(MAX30105_fire_percentage_arr,0)
////  FILL_ARR(MAX30105_temp_arr,0)
////  FILL_ARR(MAX30105_Green_arr,0)
////  FILL_ARR(MAX30105_Red_arr,0)
////  FILL_ARR(MAX30105_IR_arr,0)
//
//  for(int i = 0;i < STORAGE_LEN;i++){
//    time_arr[i] = 0;
//    Si_temp_arr[i] = 0;
//    Si_humi_arr[i] = 0;
//    MAX30105_temp_arr[i] = 0;
//    MAX30105_Green_arr[i] = 0;
//    MAX30105_Red_arr[i] = 0;
//    MAX30105_IR_arr[i] = 0;
//  }
//
//}


void setup(){

    //delay exists as the device needs to be awake to be uploaded too, used just incase an accidental upload to the MCU occurs
    pinMode(RED_LED,OUTPUT);
    digitalWrite(RED_LED,HIGH);
    //system_failure();
    delay(2000);
    interrupts();
    //only setup serial if debug is used, serial is only ever used if debugging should occur
   // #ifdef _DEBUG
    DEBUG_STREAM.begin(115200);
    while ((!DEBUG_STREAM) && (millis() < 10000)){
        // Wait max 10 sec for the CONSOLE_STREAM to open
    }
    DEBUG_PL("Starting.")
    //MODEM.debug(DEBUG_STREAM);
//}
//    #endif
    //turns off the LED to save power
    //powers the Si7021
//    pinMode(13,OUTPUT);
//    pinMode(MAG_INT,  OUTPUT);
//    pinMode(GPS_ENABLE,  OUTPUT);
//    pinMode(SARA_ENABLE,  OUTPUT);
//    pinMode(9,  OUTPUT);
//    pinMode(8,  OUTPUT);
//    digitalWrite(LED_BUILTIN, LOW);
//    digitalWrite(LED_BUILTIN, LOW);
//    digitalWrite(MAG_INT, LOW);       // we need to make this low otherwise this pin on the LSM303AGR starts leaking current
//    digitalWrite(GPS_ENABLE, LOW);    
//    digitalWrite(SARA_ENABLE, HIGH);
//    pinMode(PIN_WIRE_SDA, INPUT_PULLUP);
//    pinMode(PIN_WIRE_SCL, INPUT_PULLUP);
//    pinMode(PIN_WIRE1_SDA, INPUT_PULLUP);
//    pinMode(PIN_WIRE1_SCL, INPUT_PULLUP);
    // //Used to make sure the NB modem is powered
    // pinMode(SARA_PWR_ON,OUTPUT);
    // digitalWrite(SARA_PWR_ON,HIGH);
    //sets up external interrupt pin
//    pinMode(INTERRUPT_PIN, INPUT_PULLUP);
//    LowPower.attachInterruptWakeup(digitalPinToInterrupt(INTERRUPT_PIN),InterruptFunc,LOW);

    //random seed used for dummy data, only used if _CHECK_SENSORS is false
    //randomSeed(1);
    //setup Real Time Clock
    //set up modem, this is used to get current time for the Real Time Clock. The modem must be on before the RTC is set with it otherwise a freeze will occur
    
    setup_modem();
    //get the time from the modem
    GetTime(true);
    //shutdown the modem as it doesnt need to be used now, only needed during data transmission
    shutdown_modem(false);
    //setup everything else
    setup_sensors();
 //   setup_vars();
}
//convert the time_t day of the week value into the RTC date of time enum
RTC_DOW time_DOW2RTC_DOW(time_t input_time){
  int weekday_var = weekday(input_time);
  switch(weekday_var){
    case 1:
      return SUNDAY;
    case 2:
      return MONDAY;
    case 3:
      return TUESDAY;
    case 4:
      return WEDNESDAY;
    case 5:
      return THURSDAY;
    case 6:
      return FRIDAY;
    case 7:
      return SATURDAY;
    default:
      return NO_ALARM;
  }
}
//get RTC time and convert it into a time_t variable
time_t getRTCTime(){
  tmElements_t tmp_time;
  tmp_time.Year = RTC_getYear();
  tmp_time.Month = RTC_getMonth();
  tmp_time.Day = RTC_getDay();
  tmp_time.Hour = RTC_getHour();
  tmp_time.Minute = RTC_getMinute();
  tmp_time.Second = RTC_getSecond();
  return makeTime(tmp_time);
}
//function gets the current time, if "from_modem" is true the time is read from the NB modem (the modem must be on) and this is used to set the RTC
void GetTime(bool from_modem){
  if(from_modem){
    //get NB modems time in seconds from the last epoch (search Unix epoch)
    current_time = nbAccess.getTime();
    tmElements_t tmp_time;
    breakTime(current_time, tmp_time);
    //set RTC using the current time data, should be from the modem
    RTC_init(tmp_time.Second, tmp_time.Minute, tmp_time.Hour, tmp_time.Day,time_DOW2RTC_DOW(current_time),tmp_time.Month,tmp_time.Year);
    DEBUG_PL(tmp_time.Second);
    #ifdef _DEBUG
    
    
    DEBUG_P("Current Modem Time: ")
    DEBUG_PL((unsigned long)current_time)
    DEBUG_P("Current Clock Time: ");
    uint32_t rtc_time = getRTCTime();
    DEBUG_PL((unsigned long)(rtc_time));
    #endif
    
  }else{
    //get the epoch from the RTC
    current_time = getRTCTime();
    DEBUG_P("Current Clock Time: ")
    DEBUG_PL((unsigned long)current_time)
  }
//  //convert the epoch into the ISO 8601 format (is needed for Cumulocity to decode the time)
//  snprintf(time_string,50,"%4u-%02u-%02uT%02u:%02u:%02u",year(current_time),month(current_time),day(current_time),hour(current_time),minute(current_time),second(current_time));
//  DEBUG_PL(time_string);
}


void loop(){
  #ifdef _DEBUG
  DEBUG_STREAM.begin(115200);
  while (!DEBUG_STREAM){
        // Wait max 10 sec for the CONSOLE_STREAM to open
    }
  #endif
  DEBUG_PL("WOKE UP.")
  if(interrupt_triggered){
    DEBUG_PL("Interrupt was triggered.");
  }
  //#ifdef _DEBUG
  //digitalWrite(LED_BUILTIN, HIGH);
  //#endif
  DEBUG_PL("Gathering.")
  //gather the data from the sensors
  gatherData();
  DEBUG_PL("Getting Time.")
  //get the current time from RTC for the measurements
  GetTime(false);
  DEBUG_PL("Filling Arrays.")
  //add the current data to the storage arrays
  fillDataArrays();
  DEBUG_PL("Checking storage len.")
  
  //if arrays are now full or the sensors temperatures are above a value or the external interrupt is triggered, send data and reset arrays
  if(data_arr_ind >= STORAGE_LEN || MAX30105_temp >= MAX30105_TEMP_TRIGGER || Si_temp >= SI7021_TEMP_TRIGGER || interrupt_triggered){
    interrupt_triggered = false;
    //used to know how much infomation is in the storage arrays (used so if a premature send occurs it doesnt send data that is old)
    data_obtained = data_arr_ind;
    //reset the "first position" of the storage arrays
    data_arr_ind = 0;
    //start modem for transmission
    setup_modem();
    //as the modem is on, reset the RTCs clock with the NB modems time. Used to stop the cheap RTC on the MCU from drifting too much
    GetTime(true);
    //send the data to Cumulocity
    sendData();
    //shut down the modem to save power
    shutdown_modem(false);
   
  }
  //#ifdef _DEBUG
  //digitalWrite(LED_BUILTIN, LOW);
  //#endif
  DEBUG_PL("STARTED SLEEP.")
  //put the MCU to sleep for a period of time (is woken prematurely if the external interrupt is triggered)
  //delay(MEASUREMENT_INTERVAL_MS);
  #if _DEBUG
    DEBUG_STREAM.end();
  #endif

  sleep(MEASUREMENT_INTERVAL_MS);
}


//put the current value from the sensors into the next spot on the storage arrays
void fillDataArrays(){
  DEBUG_PL("Filling data arrays.");
  Si_temp_arr[data_arr_ind] = Si_temp;
  Si_humi_arr[data_arr_ind] = Si_humi;
  MAX30105_Green_arr[data_arr_ind] = MAX30105_Green;
  MAX30105_Red_arr[data_arr_ind] = MAX30105_Red;
  MAX30105_IR_arr[data_arr_ind] = MAX30105_IR;
  MAX30105_temp_arr[data_arr_ind] = MAX30105_temp;
  time_arr[data_arr_ind] = current_time;
  //
  #ifdef _DEBUG
  for(int i = 0;i < STORAGE_LEN;i++){
    snprintf(time_string,50,"%4u-%02u-%02uT%02u:%02u:%02u",year(time_arr[i]),month(time_arr[i]),day(time_arr[i]),hour(time_arr[i]),minute(time_arr[i]),second(time_arr[i]));
    DEBUG_PL(time_string);
  }
  #endif
  //move next position up 1
  data_arr_ind++;
}


//get the colour absorbtion data from the particle snensor
void getParticleSensorData(){
  MAX30105_Green = particleSensor.getGreen();
  MAX30105_Red = particleSensor.getRed();
  MAX30105_IR = particleSensor.getIR();
}


//gather data from sensors
void gatherData(){
  //if the sensors arent checked (_CHECK_SENSORS = false) then make dummy data
  if(!_CHECK_SENSORS){
    gatherDataDummy();
    return;
  }
  DEBUG_PL("Gathering data.")
  //setup sensors, sensors may be turned off after they were used to save power
  setup_sensors();
  //get MAX3105 temperature
  MAX30105_temp = particleSensor.readTemperature();
  // Measure Relative Humidity from the HTU21D or Si7021
  Si_humi = sensor.readHumidity();
  // Measure Temperature from the HTU21D or Si7021
  Si_temp = sensor.readTemperature();
  //get the particle sensors colour absorbtion values
  getParticleSensorData();
  //shutdown sensors to save power
  shutdownSensors();
}
//creates dummy data, only used if _CHECK_SENSORS is false, used for debug reasons
void gatherDataDummy(){
  DEBUG_PL("Gathering dummy data.")
  MAX30105_temp = dummyVal;
  humidity = dummyVal;
  tempc = dummyVal;
  Si_temp = dummyVal;
  Si_humi = dummyVal;

  MAX30105_Green = dummyVal;//random(101);
  MAX30105_Red = dummyVal;//random(101);
  MAX30105_IR = dummyVal;//random(101);
  dummyVal++;
}


//sends the data in the storage arrays to cumulocity using MQTT
void sendData(){
  DEBUG_PL("Sending data.")
  //connect to server
  //set the MQTT server
  client.setServer(mqtt_server,mqtt_port);
  //attempt connecting to the server, try a number of times before giving up
  for(int i = 0; i < MQTT_RETRIES;i++){
    client.connect(mqtt_name);//,mqtt_user,mqtt_pass);
    if(client.connected()) {
      DEBUG_PL("Connected to MQTT server and port.");
      break;
    }else{
      DEBUG_PL("Failed to connect to MQTT server and port.");
    }
  }
    //if client did not connect, exit the function
  if(!client.connected()){
    DEBUG_PL("MQTT server unavailable. Data can not be sent.");
    INDICATE_ERROR(MQTT_ERROR_PIN);
    return;
  }
  //send each data value in array
  //send every measurement intervals data up until the most recent data (the length of the array)
  for(int i = 0; i< data_obtained; i++){
    snprintf(time_string,50,"%4u-%02u-%02uT%02u:%02u:%02u",year(time_arr[i]),month(time_arr[i]),day(time_arr[i]),hour(time_arr[i]),minute(time_arr[i]),second(time_arr[i]));
    //snprintf(time_string_arr[i],50,"%s",time_string);
    MQTTSendFloat("s/us","Si7021_Temperature","Celcius","c",Si_temp_arr[i],time_string);
    MQTTSendFloat("s/us","Si7021_Humidity","Relative Humidity RH%","H",Si_humi_arr[i],time_string);
    MQTTSendInteger("s/us","MAX30107_Green_LED_Sensor","Integer","*",MAX30105_Green_arr[i],time_string);
    MQTTSendInteger("s/us","MAX30107_Red_LED_Sensor","Integer","*",MAX30105_Red_arr[i],time_string);
    MQTTSendInteger("s/us","MAX30107_Infrared_LED_Sensor","Integer","*",MAX30105_IR_arr[i],time_string);
    MQTTSendInteger("s/us","MAX_temp","Celcius","c",MAX30105_temp_arr[i],time_string);
  }
  
}
void MQTTSendString(char* thread, char* msg){
  DEBUG_PL(msg);
  msg_length = strlen(msg);
  client.publish(thread,(uint8_t*)msg,msg_length);
  //DEBUG_PL(r4x.mqttPublish(thread,(uint8_t*)msg,msg_length));

}

//send an integer to cumulocity using MQTT
//MQTT message format:  200 (code for new measurement, it will make a spot to save it based off the name of the data)
//                      name : the name of the measurement. Eg, temperature
//                      unit_name : the name of the units the data is in. Eg, Celcius
//                      data  : the measurement value, integer (because snprintf needs to have the data type)
//                      unit  : symbol of the unit. Eg, C
//                      time  : a string containing the date and time as a string in the OSI 9601 format
void MQTTSendInteger(char* thread,char* name, char* unit_name, char* unit, int data,char* data_time){
  snprintf(msg,MSG_LEN, "200,%s,%s,%ld,%s,%s\n",name,unit_name,data,unit,data_time);
  MQTTSendString(thread,msg);
}


//send a float to cumulocity using MQTT
//MQTT message format:  200 (code for new measurement, it will make a spot to save it based off the name of the data)
//                      name : the name of the measurement. Eg, temperature
//                      unit_name : the name of the units the data is in. Eg, Celcius
//                      data  : the measurement value, float (because snprintf needs to have the data type)
//                      unit  : symbol of the unit. Eg, C
//                      time  : a string containing the date and time as a string in the OSI 9601 format
void MQTTSendFloat(char* thread,char* name, char* unit_name, char* unit, float data,char* data_time){
  snprintf(msg,MSG_LEN, "200,%s,%s,%f,%s,%s\n",name,unit_name,data,unit,data_time);
  MQTTSendString(thread,msg);
}


//function that is run when the external interrupt is triggered
void InterruptFunc(){
  interrupt_triggered = true;
}
